//***************************************************************************
//
// DISK.CPP
// by Harlen C. Batagelo
//
//***************************************************************************

#if defined(WIN32)
#include <windows.h>
#pragma warning (disable:4244)	// disable bogus conversion warnings
#endif

#include <math.h>
#include <GL/gl.h>

#include "disk.h"

#define PI2 6.28318530717958647692528676655901 // PI * 2

/////////////////////////////////////////////////////////////////////////////
// Move disk to next position according to the assigned velocity
/////////////////////////////////////////////////////////////////////////////
void Disk::Move ()
{
	pos_x += inc_x;
	pos_y += inc_y;

	if (pos_x > width - radius - 3) inc_x = -inc_x;
	if (pos_y > height - radius - 3) inc_y = -inc_y;
	if (pos_x < 3 + radius) inc_x = -inc_x;
	if (pos_y < 3 + radius) inc_y = -inc_y;
} // Disk::Move

/////////////////////////////////////////////////////////////////////////////
// Draw a filled disk in light red with an outline in dark red
/////////////////////////////////////////////////////////////////////////////
void Disk::DrawDisk ()
{
	glColor3f (1.0, 0.0, 0.0);
	glBegin (GL_TRIANGLE_FAN);
		glVertex3f (pos_x, pos_y, 0);
		for(double u = 0; u <= PI2; u += PI2 / sides)
			glVertex3f (pos_x + (radius*cos(u)), 
						pos_y + (radius*sin(u)), 0);
	glEnd ();

	glColor3f (0.5, 0.0, 0.0);
	glLineWidth (2.0);
	glBegin (GL_LINE_LOOP);
		for(u = 0; u <= PI2; u += PI2 / sides)
			glVertex3f (pos_x + (radius*cos(u)), 
						pos_y + (radius*sin(u)), 0);
	glEnd ();
} // Disk::DrawDisk
