//***************************************************************************
//
// DISK.H
// by Harlen C. Batagelo
//
//***************************************************************************

#ifndef _DISK_H
#define _DISK_H

/////////////////////////////////////////////////////////////////////////////
// Define a disk
/////////////////////////////////////////////////////////////////////////////
class Disk
{
private:
	float pos_x, pos_y;		// disk position
	float inc_x, inc_y;		// velocity
	float radius;			// radius of the disk
	float sides;			// number of sides
	float width, height;	// width and height of the moving plane

	// TBV date
	int has_tbv;			// TRUE if this object has a TBV
	float tbv_x, tbv_y;		// TBV position
	long start_date;		// TBV start date (absolute date)
	long expiration_date;   // TBV expiration date (absolute date)
	float last_x, last_y;	// disk position at the last time this disk was drawn
	float TBV_radius;		// radius of the TBV

public:
	Disk ()
	{
		radius = 0;			// default settings
		sides = 3;
		pos_x = pos_y = 0;
		inc_x = inc_y = 1.0; 
		width = height = 0;
	}

	inline void SetPos (float x, float y) { pos_x = x; pos_y = y; }
	inline void SetRadius (float r) { radius = r; }
	inline void SetSides (float s) { sides = s; }
	inline void SetVelocity (float v) { inc_x = inc_y = v; }
	inline void SetBounds (float w, float h) { width = w; height = h; }
	inline void SetStartDate (long date) { start_date = date; }
	inline void SetExpDate (long date) { expiration_date = date; }
	inline void SetTBVX (float x) { tbv_x = x; }
	inline void SetTBVY (float y) { tbv_y = y; }
	inline void SetLastX (float x) { last_x = x; }
	inline void SetLastY (float y) { last_y = y; }
	inline void SetTBVState (int s) { has_tbv = s; }
	inline void SetTBVRadius (int r) { TBV_radius = r; }

	inline float GetX () { return pos_x; }
	inline float GetY () { return pos_y; }
	inline float GetRadius () { return radius; }
	inline float GetSides () { return sides; }
	inline float GetVelocityX () { return inc_x; }
	inline float GetVelocityY () { return inc_y; }
	inline float GetBoundWidth () { return width; }
	inline float GetBoundHeight () { return height; }
	inline long GetStartDate () { return start_date; }
	inline long GetExpDate () { return expiration_date; }
	inline long GetTBVX () { return tbv_x; }
	inline long GetTBVY () { return tbv_y; }
	inline float GetLastX () { return last_x; }
	inline float GetLastY () { return last_y; }
	inline int HasTBV () { return has_tbv; }
	inline float GetTBVRadius () { return TBV_radius; }

	void Move ();		// move disk to next position
	void DrawDisk ();	// draw a filled disk
};

#endif