//***************************************************************************
//
// MAIN.CPP
// by Harlen C. Batagelo
//
//***************************************************************************

#pragma warning(disable:4786)
#include <set>
#include <list>

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include <GL/glut.h>

#include "fps.h"
#include "disk.h"
#include "poly2d.h"
#include "bitset.h"

int flagInsideObj = 0;

#define NHOC			// uncomment this to enable occlusion culling

#define WIDTH  256
#define HEIGHT 256
#define N_DISKS_DYN 300 
#define N_DISKS_STC 0

#define MIN_POT_OCCLUDER 2.0f

#define ROUND(x) ((int)((x)+.5)) 
#define MK_OFFSET(x,y) (((y)*WIDTH)+(x))
#define MK_OFFSET_CLIPXY(x,y) ((((y<1)?1:((y>HEIGHT-2)?HEIGHT-2:y))*WIDTH)+ \
								((x<1)?1:((x>WIDTH-2)?WIDTH-2:x)))
#define MAX(a,b) ((a>b)?a:b)
#define SQUARE(a,b) (sqrt((a)*(a)+(b)*(b)))
#define RAD(x) ((x) * 0.017453293)

#if !defined(WIN32)
enum BOOL { FALSE = 0, TRUE };
#endif

float *readBuffer;
float *readBufferR, *readBufferG, *readBufferB;
float overwrite_max, overwrite_cur, overwrite_min;

int n_disks;
Disk *disks;

typedef std::set<int> set_of_ids;

typedef struct
{
	long frame;
	float fps;
	long visobj;
	long occluders;
	float otherValue;
} FRAME_INFO;

typedef std::list<FRAME_INFO*> log_list;
log_list *logfile = NULL;	// Log of recorded FPS

Poly2D poly;
GLUquadricObj *qobj;

set_of_ids I[WIDTH*HEIGHT];	// IDs matrix
BitSet H(WIDTH*HEIGHT);		// occlusion matrix
BitSet O(WIDTH*HEIGHT);		// occluders matrix
BitSet O_static(WIDTH*HEIGHT);		// occluders matrix for static objects
char PVS[N_DISKS_DYN+N_DISKS_STC];	// PVS based on IDs matrix

set_of_ids T[WIDTH*HEIGHT]; // TBVs matrix

float rot	= 45.0;			// rotation angle (in degrees)
float vPosX = 41.0, 		// viewpoint position
	  vPosY = 183.0;	
float fov	= (float) 55;	// field of view (in degrees) = 35mm camera

// distance of the far plane from the viewpoint
float zFar	= sqrt(WIDTH*WIDTH + HEIGHT*HEIGHT);

// corner points that define the view frustum
float corner1X, corner1Y, corner2X, corner2Y; 
float tCorner1X, tCorner1Y, tCorner2X, tCorner2Y; // transformed points

// coefficients of the left and right clipping lines
float al, bl, cl, ar, br, cr;

// used by startTraversal() to speed up the view frustum traversal
static int goUp = TRUE, goDown = TRUE, goLeft = TRUE, goRight = TRUE;

static int occluders;
static BOOL show_data = TRUE, is_dynamic = TRUE;
static long current_date = 0;

void upd_vis (int ct, int draw_occ);
void keyboard (unsigned char key, int x, int y);

// used to store a color for each object
static float R[N_DISKS_DYN+N_DISKS_STC], 
			 G[N_DISKS_DYN+N_DISKS_STC], 
			 B[N_DISKS_DYN+N_DISKS_STC];

void WriteLogFile (void)
{
	FILE *stream;
	double sum1 = 0, sum2 = 0, sum3 = 0, sum4 = 0;
	long quantity = 0;

	if (!(stream = fopen("fps.log", "wt")))
		return;

	fprintf(stream, "Frame number | FPS | otherValue\n\n");

	for (log_list::iterator i = logfile->begin(); i != logfile->end(); ++i)
	{
		sum1 += (*i)->fps;
		sum2 += (*i)->visobj;
		sum3 += (*i)->occluders;
		sum4 += (*i)->otherValue;
		quantity++;
		fprintf(stream, "%d    %.2f    %.2f\n", (*i)->frame, 
				(*i)->fps, (*i)->otherValue);
	}

	fprintf(stream, "Mean: %.2f FPS     %.2f visObjs     %.2f occluders     %.2f tightness\n", 
			sum1/quantity, sum2/quantity, sum3/quantity, sum4/quantity);

	fclose (stream);
}

void AddToLog (float FPS, long visobj, long occluders, float otherValue)
{
	if (!logfile)
		logfile = new log_list;

	FRAME_INFO *info;
	info = new FRAME_INFO;
	info->fps = FPS;
	info->frame = current_date;
	info->visobj = visobj;
	info->occluders = occluders;
	info->otherValue = otherValue;

	logfile->insert (logfile->end(), info);
}

/////////////////////////////////////////////////////////////////////////////
// Assign a TBV to a given object
/////////////////////////////////////////////////////////////////////////////
void insertTBV (int disk)
{
	int x0 = ROUND (disks[disk].GetTBVX ());
	int y0 = ROUND ((HEIGHT-1) - disks[disk].GetTBVY ());
	int radius = ROUND (disks[disk].GetRadius () +
				 (float) (disks[disk].GetExpDate () - 
				 disks[disk].GetStartDate ()) * 
				 SQUARE (disks[disk].GetVelocityX(),
					  disks[disk].GetVelocityY()));

	if (radius > SQUARE(WIDTH/5.0, HEIGHT/5.0))
	{
		radius = SQUARE(WIDTH/5.0, HEIGHT/5.0);

		// set expiration date based on the new radius
		disks[disk].SetExpDate (ROUND (((radius - disks[disk].GetRadius ()) /
								SQUARE(disks[disk].GetVelocityX(),
								disks[disk].GetVelocityY())) + 
								disks[disk].GetStartDate()));

	}

	// radius must be smaller than the distance from the
	// disk center to the viewer
	int r = ROUND (SQUARE(vPosX-x0, ((HEIGHT-1)-vPosY)-y0));
	if (radius >= r)
	{
		radius = r - 1;

		if (radius < disks[disk].GetRadius ())
			radius = disks[disk].GetRadius ();

		// set expiration date based on the new radius
		disks[disk].SetExpDate (ROUND (((radius - disks[disk].GetRadius ()) /
								SQUARE(disks[disk].GetVelocityX(),
								disks[disk].GetVelocityY())) + 
								disks[disk].GetStartDate()));
	}

	disks[disk].SetTBVRadius (radius);

	int d = 3 - (radius << 1);
	int dh = 6, dd = 10 - (radius << 2);

	for (int ix = 0, iy = radius; ix <= iy; ix++)
	{
		T[MK_OFFSET_CLIPXY (x0+ix, y0+iy)].insert (disk);
		T[MK_OFFSET_CLIPXY (x0+iy, y0+ix)].insert (disk);
		T[MK_OFFSET_CLIPXY (x0+ix, y0-iy)].insert (disk);
		T[MK_OFFSET_CLIPXY (x0+iy, y0-ix)].insert (disk);
		T[MK_OFFSET_CLIPXY (x0-ix, y0+iy)].insert (disk);
		T[MK_OFFSET_CLIPXY (x0-iy, y0+ix)].insert (disk);
		T[MK_OFFSET_CLIPXY (x0-ix, y0-iy)].insert (disk);
		T[MK_OFFSET_CLIPXY (x0-iy, y0-ix)].insert (disk); 

		if (d >= 0)
		{
			iy--; // diagonal increase
			d += dd;
			dd += 8;
		}
		else
		{
			d += dh; // horizontal increase
			dd += 4;
		}
		dh += 4;
	}
} // insertTBV

/////////////////////////////////////////////////////////////////////////////
// Remove the TBV of a given object
/////////////////////////////////////////////////////////////////////////////
void eraseTBV (int disk)
{
	int x0 = ROUND (disks[disk].GetTBVX ());
	int y0 = ROUND ((HEIGHT-1) - disks[disk].GetTBVY ());
	int radius = disks[disk].GetTBVRadius ();

	int d = 3 - (radius << 1);
	int dh = 6, dd = 10 - (radius << 2);

	for (int ix = 0, iy = radius; ix <= iy; ix++)
	{
		T[MK_OFFSET_CLIPXY (x0+ix, y0+iy)].erase (disk);
		T[MK_OFFSET_CLIPXY (x0+iy, y0+ix)].erase (disk);
		T[MK_OFFSET_CLIPXY (x0+ix, y0-iy)].erase (disk);
		T[MK_OFFSET_CLIPXY (x0+iy, y0-ix)].erase (disk);
		T[MK_OFFSET_CLIPXY (x0-ix, y0+iy)].erase (disk);
		T[MK_OFFSET_CLIPXY (x0-iy, y0+ix)].erase (disk);
		T[MK_OFFSET_CLIPXY (x0-ix, y0-iy)].erase (disk);
		T[MK_OFFSET_CLIPXY (x0-iy, y0-ix)].erase (disk); 

		if (d >= 0)
		{
			iy--; // diagonal increase
			d += dd;
			dd += 8;
		}
		else
		{
			d += dh; // horizontal increase
			dd += 4;
		}
		dh += 4;
	}
} // eraseTBV

/////////////////////////////////////////////////////////////////////////////
// Add to the PVS the objects associated with TBVs of the given cell 
/////////////////////////////////////////////////////////////////////////////
inline void add2PVSTBV (int x, int y)
{
	static int PVSTBV[N_DISKS_DYN];
	long i = 0;

	// iterate over the list of objects
	set_of_ids::iterator setIt;
	for (setIt = T[MK_OFFSET(x, y)].begin(); 
		setIt != T[MK_OFFSET(x, y)].end(); setIt++)
	{
		upd_vis (*setIt, 1);

		if(O[MK_OFFSET(vPosX, (HEIGHT-1)-vPosY)])
			flagInsideObj = 1;

		PVSTBV[i++] = *setIt;
	}

	for(long ct = 0; ct < i; ct++)
	{
		eraseTBV (PVSTBV[ct]);
		disks[PVSTBV[ct]].SetTBVState (0);
	}
} // add2PVSTBV

/////////////////////////////////////////////////////////////////////////////
// Add the objects associated with the given cell to the PVS
/////////////////////////////////////////////////////////////////////////////
inline void add2PVS (int x, int y, BOOL hid)
{
	if (H[MK_OFFSET(x, y)])
		return;

	// iterate over the list of objects
	set_of_ids::iterator setIt;
	for (setIt = I[MK_OFFSET(x, y)].begin(); 
		 setIt != I[MK_OFFSET(x, y)].end(); setIt++)
	{
		if (!PVS[*setIt])
			PVS[*setIt] = 1;
	}

	if(hid)
		H.set(MK_OFFSET(x, y));

	add2PVSTBV (x, y);
} // add2PVS

/////////////////////////////////////////////////////////////////////////////
// Draw occluder shaft and sets hidden cells
/////////////////////////////////////////////////////////////////////////////
void drawShaft (void)
{
	int ledge[2000], redge[2000]; // buffers to store x values
	int maxv, minv; // top and bottom vertices

	// clip shaft
	poly.ClipRect (1, 1, WIDTH-2, HEIGHT-1);

	// determine top and bottom vertices
	float miny =  999999.0,	maxy = -999999.0;
	for (int ct = 0; ct < poly.GetSize (); ct++)
	{
		float this_y = poly.GetY (ct);
		if (this_y < miny) { minv = ct; miny = this_y; }
		if (this_y > maxy) { maxv = ct; maxy = this_y; }
	}

	// determine leftmost and rightmost vertices
	float minx =  999999.0, maxx = -999999.0;
	for (ct = 0; ct < poly.GetSize (); ct++)
	{
		int this_x = poly.GetX (ct);
		if (this_x < minx) { minx = this_x; }
		if (this_x > maxx) { maxx = this_x; }
	}

	int iminy = ROUND (miny); // convert to integer
	int imaxy = ROUND (maxy);

	if (iminy == imaxy) return; // nothing to draw

	// scan out the left edge
	int leftv = minv; // x1, y1
	do
	{
		int next = leftv - 1; // x2, y2
		if (next < 0) next = poly.GetSize () - 1;

		float x1 = poly.GetX (leftv), y1 = poly.GetY (leftv),
			  x2 = poly.GetX (next),  y2 = poly.GetY (next);

		if (y1 < y2)
		{
			float mx = (x2 - x1) / (y2 - y1),
				  x = x1;

			for (int ct = ROUND (y1); ct < ROUND (y2); ct++, x += mx)
				ledge[ct] = ROUND (x);
		}
		leftv--; // next vertex
		if (leftv < 0) leftv = poly.GetSize () - 1;
	} while (leftv != maxv);

	// scan out the right edge
	int rightv = minv; // x1, y1
	do
	{
		int next = (rightv + 1) % poly.GetSize (); // x2, y2

		float x1 = poly.GetX (rightv), y1 = poly.GetY (rightv),
			  x2 = poly.GetX (next),   y2 = poly.GetY (next);

		if (y1 < y2)
		{
			float mx = (x2 - x1) / (y2 - y1),
	 			  x = x1;

			for (int ct = ROUND (y1); ct < ROUND (y2); ct++, x += mx)
				redge[ct] = ROUND (x);
		}
		rightv = (rightv + 1) % poly.GetSize (); // next vertex
	} while(rightv != maxv);

	// draw the horizontal lines
	for (ct = iminy; ct < imaxy; ct++)
	{
		int x1 = ledge[ct], x2 = redge[ct];
		if (x2 < x1) 
		{
			float tmp = x1;
			x1 = x2;
			x2 = tmp;
		}

		H.set ((ct * WIDTH)+x1, (x2-x1)+1);
	}
} // drawShaft

/////////////////////////////////////////////////////////////////////////////
// Test whether the cell (x,y) is inside the view frustum
/////////////////////////////////////////////////////////////////////////////
inline BOOL isInsideVF (float x, float y)
{
	if (al*x+bl*y+cl < 0 && ar*x+br*y+cr < 0)
		return TRUE;

	return FALSE;
} // isInsideVF

/////////////////////////////////////////////////////////////////////////////
// Test whether the cell (x,y) is a valid occluder
/////////////////////////////////////////////////////////////////////////////
inline BOOL isValidOccluder (int x, int y)
{
	if (isInsideVF (x, (HEIGHT-1)-y))
		if (O[MK_OFFSET(x, y)] || H[MK_OFFSET(x, y)])
			return TRUE;

	return FALSE;
} // isValideOccluder

/////////////////////////////////////////////////////////////////////////////
// Return a quantity that estimates the potential of the occluder
// The greater is the value, the better is the occluder
/////////////////////////////////////////////////////////////////////////////
inline float occlusionPotential (int area, int x_dist, int y_dist)
{
	return (area*area*5000) / pow (SQUARE(x_dist, y_dist), 2);
} // occlusionPotential

/////////////////////////////////////////////////////////////////////////////
// Try to extend the occluder and rasterizes an occluder shaft
/////////////////////////////////////////////////////////////////////////////
inline void setOccluderShaft (int x, int y, int x_dist, int y_dist)
{
	int vPosY_flipped = (HEIGHT-1)-vPosY;

	// determine viewpoint region relative to the occluder
	int x_region, y_region; // 0 = left/top, 1 = middle, 2 = right/bottom

	//  0,0 | 1,0 | 2,0
	//      |     |
	// -----+-----+-----
	//  0,1 |occlu| 2,1
	//      |der  |
	// -----+-----+-----
	//  0,2 | 1,2 | 2,2
	//      |     |

	(vPosX < x) ? x_region = 0 : ((vPosX > x) ? x_region = 2 : x_region = 1);
	(vPosY_flipped < y) ? y_region = 0 : 
		((vPosY_flipped > y) ? y_region = 2 : y_region = 1);

	int x1 = x, y1 = y, x2 = x, y2 = y;
	BOOL first_passed = FALSE;
	int shaft_area = 0; // size of the shaft in cells

	switch (x_region)
	{
	case 0:
		switch (y_region)
		{
		case 0:
			// extension heuristic
			//
			//            -4th->
			//          /|\
			//          3rd (only if 1st didn't work)
			//           |
			//    <-1st-[*] occluder
			//    |
			//   2nd
			//   \|/

			// try to extend this occluder to the left
			// (beginning from the start point)
			while (isValidOccluder(x1-1, y))
			{	
				x1--;
				first_passed = TRUE;
				add2PVS (x1, y, TRUE);
				shaft_area++;
			}

			if (x1 <= vPosX) // left extension has passed the viewer
			{
				// try to extend to the right
				while (isValidOccluder(x2+1, y))
				{	
					x2++;
					add2PVS (x2, y, TRUE);
					shaft_area++;
				}

				if (occlusionPotential (shaft_area, x_dist, y_dist) < MIN_POT_OCCLUDER) 
					return; // shaft too small

				// vertices of the occlusion polygon
				poly.SetXY (0, x1, y1);
				poly.SetXY (1, x2, y1);
				poly.SetXY (2, vPosX+(x2-vPosX)*1000, vPosY_flipped+(y1-vPosY_flipped)*1000);
				poly.SetXY (3, vPosX+(x1-vPosX)*1000, vPosY_flipped+(y1-vPosY_flipped)*1000);
				poly.SetSize (4);

				drawShaft ();

				return;
			}

			// try to extend to the bottom
			while (isValidOccluder(x1, y2+1))
			{
				y2++;
				add2PVS (x1, y2, TRUE);
				shaft_area++;
			}

			// try to extend to the top
			// (beginning from the start point)
			if (!first_passed)
			{
				while (isValidOccluder(x, y1-1))
				{
					y1--;
					add2PVS (x, y1, TRUE);
					shaft_area++;
				}
			}

			if (y1 <= vPosY_flipped) // top extension has passed the viewer
			{
				if (occlusionPotential (shaft_area, x_dist, y_dist) < MIN_POT_OCCLUDER) 
					return; // shaft too small

				// vertices of the occlusion polygon
				poly.SetXY (0, x1, y1);
				poly.SetXY (1, vPosX+(x1-vPosX)*1000, vPosY_flipped+(y1-vPosY_flipped)*1000);
				poly.SetXY (2, vPosX+(x1-vPosX)*1000, vPosY_flipped+(y2-vPosY_flipped)*1000);
				poly.SetXY (3, x1, y2);
				poly.SetSize (4);

				drawShaft ();

				return;
			}

			// try to extend to the right
			while (isValidOccluder(x2+1, y1))
			{	
				x2++;
				add2PVS (x2, y1, TRUE);
				shaft_area++;
			}

			if (occlusionPotential (shaft_area, x_dist, y_dist) < MIN_POT_OCCLUDER) 
				return; // shaft too small

			// vertices of the occlusion polygon
			poly.SetXY (0, x1, y1);
			poly.SetXY (1, x2, y1);
			poly.SetXY (2, vPosX+(x2-vPosX)*1000, vPosY_flipped+(y1-vPosY_flipped)*1000);
			poly.SetXY (3, vPosX+(x1-vPosX)*1000, vPosY_flipped+(y2-vPosY_flipped)*1000);
			poly.SetXY (4, x1, y2);
			poly.SetSize (5);

			drawShaft ();

			break;

		case 1:
			// extension heuristic
			//
			//   /|\
			//   1st
			//    |
			//   [*] occluder
			//    |
			//   2nd
			//   \|/

			// extend to the top
			while (isValidOccluder(x, y1-1))
			{	
				y1--;
				add2PVS (x, y1, TRUE);
				shaft_area++;
			}

			// extend to the bottom
			while (isValidOccluder(x, y2+1))
			{	
				y2++;
				add2PVS (x, y2, TRUE);
				shaft_area++;
			}

			if (occlusionPotential (shaft_area, x_dist, y_dist) < MIN_POT_OCCLUDER) 
				return; // shaft too small

			// vertices of the occlusion polygon
			poly.SetXY (0, x, y1);
			poly.SetXY (1, x, y2);
			poly.SetXY (2, vPosX+(x-vPosX)*1000, vPosY_flipped+(y2-vPosY_flipped)*1000);
			poly.SetXY (3, vPosX+(x-vPosX)*1000, vPosY_flipped+(y1-vPosY_flipped)*1000);
			poly.SetSize (4);

			drawShaft ();

			break;

		case 2:
			// extension heuristic
			//
			//   
			//   /|\
			//   4th 
			//    |   occluder
			//    <-3rd-[*] (do 3rd only if 1st didn't work)
			//           |
			//          1st
			//          \|/
			//            -2nd->

			// try to extend this occluder to the bottom
			// (beginning from the start point)
			while (isValidOccluder(x, y2+1))
			{	
				y2++;
				first_passed = TRUE;
				add2PVS (x, y2, TRUE);
				shaft_area++;
			}

			if (y2 >= vPosY_flipped) // bottom extension has passed the viewer
			{
				// try to extend to the top
				while (isValidOccluder(x, y1-1))
				{	
					y1--;
					add2PVS (x, y1, TRUE);
					shaft_area++;
				}

				if (occlusionPotential (shaft_area, x_dist, y_dist) < MIN_POT_OCCLUDER) 
					return; // shaft too small

				// vertices of the occlusion polygon
				poly.SetXY (0, x, y1);
				poly.SetXY (1, x, y2);
				poly.SetXY (2, vPosX+(x-vPosX)*1000, vPosY_flipped+(y2-vPosY_flipped)*1000);
				poly.SetXY (3, vPosX+(x-vPosX)*1000, vPosY_flipped+(y1-vPosY_flipped)*1000);
				poly.SetSize (4);

				drawShaft ();

				return;
			}

			// try to extend to the right
			while (isValidOccluder(x2+1, y2))
			{
				x2++;
				add2PVS (x2, y2, TRUE);
				shaft_area++;
			}

			// try to extend to the left
			// (beginning from the start point)
			if (!first_passed)
			{
				while (isValidOccluder(x1-1, y))
				{
					x1--;
					add2PVS (x1, y, TRUE);
					shaft_area++;
				}
			}

			if (x1 <= vPosX) // left extension has passed the viewer
			{
				if (occlusionPotential (shaft_area, x_dist, y_dist) < MIN_POT_OCCLUDER) 
					return; // shaft too small

				// vertices of the occlusion polygon
				poly.SetXY (0, x1, y);
				poly.SetXY (1, x2, y);
				poly.SetXY (2, vPosX+(x2-vPosX)*1000, vPosY_flipped+(y-vPosY_flipped)*1000);
				poly.SetXY (3, vPosX+(x1-vPosX)*1000, vPosY_flipped+(y-vPosY_flipped)*1000);
				poly.SetSize (4);

				drawShaft ();

				return;
			}

			// try to extend to the top
			while (isValidOccluder(x1, y1-1))
			{	
				y1--;
				add2PVS (x1, y1, TRUE);
				shaft_area++;
			}

			if (occlusionPotential (shaft_area, x_dist, y_dist) < MIN_POT_OCCLUDER) 
				return; // shaft too small

			// vertices of the occlusion polygon
			poly.SetXY (0, x1, y1);
			poly.SetXY (1, x1, y2);
			poly.SetXY (2, x2, y2);
			poly.SetXY (3, vPosX+(x2-vPosX)*1000, vPosY_flipped+(y2-vPosY_flipped)*1000);
			poly.SetXY (4, vPosX+(x1-vPosX)*1000, vPosY_flipped+(y1-vPosY_flipped)*1000);
			poly.SetSize (5);

			drawShaft ();

			break;
		}
		break;

	case 1:
		switch (y_region)
		{
		case 0:
			// extension heuristic
			//
			//  <-1st-[*]-2nd->
			//     occluder

			// extend to the left
			while (isValidOccluder(x1-1, y))
			{	
				x1--;
				add2PVS (x1, y, TRUE);
				shaft_area++;
			}

			// extend to the right
			while (isValidOccluder(x2+1, y))
			{	
				x2++;
				add2PVS (x2, y, TRUE);
				shaft_area++;
			}

			if (occlusionPotential (shaft_area, x_dist, y_dist) < MIN_POT_OCCLUDER) 
				return; // shaft too small

			// vertices of the occlusion polygon
			poly.SetXY (0, x1, y);
			poly.SetXY (1, x2, y);
			poly.SetXY (2, vPosX+(x2-vPosX)*1000, vPosY_flipped+(y-vPosY_flipped)*1000);
			poly.SetXY (3, vPosX+(x1-vPosX)*1000, vPosY_flipped+(y-vPosY_flipped)*1000);
			poly.SetSize (4);

			drawShaft ();

			break;

		case 1:
			// extension heuristic
			//
			//   /|\
			//   1st
			//    |
			//   [*] occluder
			//    |
			//   2nd
			//   \|/

			// extend to the top
			while (isValidOccluder(x, y1-1))
			{	
				y1--;
				add2PVS (x, y1, TRUE);
				shaft_area++;
			}

			// extend to the bottom
			while (isValidOccluder(x, y2+1))
			{	
				y2++;
				add2PVS (x, y2, TRUE);
				shaft_area++;
			}

			if (occlusionPotential (shaft_area, x_dist, y_dist) < MIN_POT_OCCLUDER) 
				return; // shaft too small

			// vertices of the occlusion polygon
			poly.SetXY (0, x, y1);
			poly.SetXY (1, x, y2);
			poly.SetXY (2, vPosX+(x-vPosX)*1000, vPosY_flipped+(y2-vPosY_flipped)*1000);
			poly.SetXY (3, vPosX+(x-vPosX)*1000, vPosY_flipped+(y1-vPosY_flipped)*1000);
			poly.SetSize (4);

			drawShaft ();

			break;

		case 2:
			// extension heuristic
			//
			//  <-1st-[*]-2nd->
			//     occluder

			// extend to the left
			while (isValidOccluder(x1-1, y))
			{	
				x1--;
				add2PVS (x1, y, TRUE);
				shaft_area++;
			}

			// extend to the right
			while (isValidOccluder(x2+1, y))
			{	
				x2++;
				add2PVS (x2, y, TRUE);
				shaft_area++;
			}

			if (occlusionPotential (shaft_area, x_dist, y_dist) < MIN_POT_OCCLUDER) 
				return; // shaft too small

			// vertices of the occlusion polygon
			poly.SetXY (0, x1, y);
			poly.SetXY (1, x2, y);
			poly.SetXY (2, vPosX+(x2-vPosX)*1000, vPosY_flipped+(y-vPosY_flipped)*1000);
			poly.SetXY (3, vPosX+(x1-vPosX)*1000, vPosY_flipped+(y-vPosY_flipped)*1000);
			poly.SetSize (4);

			drawShaft ();

			break;
		}

	case 2:
		switch (y_region)
		{
		case 0:
			// extension heuristic
			//
			//     <-2nd- 
			//          /|\
			//          1st 
			//           |
			// occluder [*]-3rd-> (only if 1st didn't work)
			//                  |
			//                 4th
			//                 \|/

			// extend to the top
			// (beginning from the start point)
			while (isValidOccluder(x, y1-1))
			{	
				y1--;
				first_passed = TRUE;
				add2PVS (x, y1, TRUE);
				shaft_area++;
			}

			if (y1 <= vPosY_flipped) // top extension has passed the viewer
			{
				// try to extend to the bottom
				while (isValidOccluder(x, y2+1))
				{	
					y2++;
					add2PVS (x, y2, TRUE);
					shaft_area++;
				}

				if (occlusionPotential (shaft_area, x_dist, y_dist) < MIN_POT_OCCLUDER) 
					return; // shaft too small

				// vertices of the occlusion polygon
				poly.SetXY (0, x, y1);
				poly.SetXY (1, x, y2);
				poly.SetXY (2, vPosX+(x-vPosX)*1000, vPosY_flipped+(y2-vPosY_flipped)*1000);
				poly.SetXY (3, vPosX+(x-vPosX)*1000, vPosY_flipped+(y1-vPosY_flipped)*1000);
				poly.SetSize (4);

				drawShaft ();

				return;
			}

			// try to extend to the left
			while (isValidOccluder(x1-1, y1))
			{
				x1--;
				add2PVS (x1, y1, TRUE);
				shaft_area++;
			}

			// try to extend to the right
			// (beginning from the start point)
			if (!first_passed)
			{
				while (isValidOccluder(x2+1, y))
				{
					x2++;
					add2PVS (x2, y, TRUE);
					shaft_area++;
				}
			}

			if (x2 >= vPosX) // right extension has passed the viewer
			{
				if (occlusionPotential (shaft_area, x_dist, y_dist) < MIN_POT_OCCLUDER) 
					return; // shaft too small

				// vertices of the occlusion polygon
				poly.SetXY (0, x1, y);
				poly.SetXY (1, x2, y);
				poly.SetXY (2, vPosX+(x2-vPosX)*1000, vPosY_flipped+(y-vPosY_flipped)*1000);
				poly.SetXY (3, vPosX+(x1-vPosX)*1000, vPosY_flipped+(y-vPosY_flipped)*1000);
				poly.SetSize (4);

				drawShaft ();

				return;
			}

			// try to extend to the bottom
			while (isValidOccluder(x2, y2+1))
			{	
				y2++;
				add2PVS (x2, y2, TRUE);
				shaft_area++;
			}

			if (occlusionPotential (shaft_area, x_dist, y_dist) < MIN_POT_OCCLUDER) 
				return; // shaft too small

			// vertices of the occlusion polygon
			poly.SetXY (0, x1, y1);
			poly.SetXY (1, x2, y1);
			poly.SetXY (2, x2, y2);
			poly.SetXY (3, vPosX+(x2-vPosX)*1000, vPosY_flipped+(y2-vPosY_flipped)*1000);
			poly.SetXY (4, vPosX+(x1-vPosX)*1000, vPosY_flipped+(y1-vPosY_flipped)*1000);
			poly.SetSize (5);

			drawShaft ();

			break;

		case 1:
			// extension heuristic
			//
			//   /|\
			//   1st
			//    |
			//   [*] occluder
			//    |
			//   2nd
			//   \|/

			// extend to the top
			while (isValidOccluder(x, y1-1))
			{	
				y1--;
				add2PVS (x, y1, TRUE);
				shaft_area++;
			}

			// extend to the bottom
			while (isValidOccluder(x, y2+1))
			{	
				y2++;
				add2PVS (x, y2, TRUE);
				shaft_area++;
			}

			if (occlusionPotential (shaft_area, x_dist, y_dist) < MIN_POT_OCCLUDER) 
				return; // shaft too small

			// vertices of the occlusion polygon
			poly.SetXY (0, x, y1);
			poly.SetXY (1, x, y2);
			poly.SetXY (2, vPosX+(x-vPosX)*1000, vPosY_flipped+(y2-vPosY_flipped)*1000);
			poly.SetXY (3, vPosX+(x-vPosX)*1000, vPosY_flipped+(y1-vPosY_flipped)*1000);
			poly.SetSize (4);

			drawShaft ();

			break;

		case 2:
			// extension heuristic
			//
			//                 /|\
			//                 2th 
			//                  |
			// occluder [*]-1st->
			//           |
			//          3rd (only if 1st didn't work)
			//          \|/
			//     <-4th-

			// try to extend this occluder to the right
			// (beginning from the start point)
			while (isValidOccluder(x2+1, y))
			{	
				x2++;
				first_passed = TRUE;
				add2PVS (x2, y, TRUE);
				shaft_area++;
			}

			if (x2 >= vPosX) // right extension has passed the viewer
			{
				// try to extend to the left
				while (isValidOccluder(x1-1, y))
				{	
					x1--;
					add2PVS (x1, y, TRUE);
					shaft_area++;
				}

				if (occlusionPotential (shaft_area, x_dist, y_dist) < MIN_POT_OCCLUDER) 
					return; // shaft too small

				// vertices of the occlusion polygon
				poly.SetXY (0, x1, y);
				poly.SetXY (1, x2, y);
				poly.SetXY (2, vPosX+(x2-vPosX)*1000, vPosY_flipped+(y-vPosY_flipped)*1000);
				poly.SetXY (3, vPosX+(x1-vPosX)*1000, vPosY_flipped+(y-vPosY_flipped)*1000);
				poly.SetSize (4);

				drawShaft ();

				return;
			}

			// try to extend to the top
			while (isValidOccluder(x2, y1-1))
			{
				y1--;
				add2PVS (x2, y1, TRUE);
				shaft_area++;
			}

			// try to extend to the bottom
			// (beginning from the start point)
			if (!first_passed)
			{
				while (isValidOccluder(x, y2+1))
				{
					y2++;
					add2PVS (x, y2, TRUE);
					shaft_area++;
				}
			}

			if (y2 >= vPosY_flipped) // bottom extension has passed the viewer
			{
				if (occlusionPotential (shaft_area, x_dist, y_dist) < MIN_POT_OCCLUDER) 
					return; // shaft too small

				// vertices of the occlusion polygon
				poly.SetXY (0, x, y1);
				poly.SetXY (1, x, y2);
				poly.SetXY (2, vPosX+(x-vPosX)*1000, vPosY_flipped+(y2-vPosY_flipped)*1000);
				poly.SetXY (3, vPosX+(x-vPosX)*1000, vPosY_flipped+(y1-vPosY_flipped)*1000);
				poly.SetSize (4);

				drawShaft ();

				return;
			}

			// try to extend to the left
			while (isValidOccluder(x1-1, y2))
			{	
				x1--;
				add2PVS (x1, y2, TRUE);
				shaft_area++;
			}

			if (occlusionPotential (shaft_area, x_dist, y_dist) < MIN_POT_OCCLUDER) 
				return; // shaft too small

			// vertices of the occlusion polygon
			poly.SetXY (0, x1, y2);
			poly.SetXY (1, x2, y2);
			poly.SetXY (2, x2, y1);
			poly.SetXY (3, vPosX+(x2-vPosX)*1000, vPosY_flipped+(y1-vPosY_flipped)*1000);
			poly.SetXY (4, vPosX+(x1-vPosX)*1000, vPosY_flipped+(y2-vPosY_flipped)*1000);
			poly.SetSize (5);

			drawShaft ();

			break;
		}
		break;
	}
} // setOccluderShaft

/////////////////////////////////////////////////////////////////////////////
// Look for occluders and occludees. Return TRUE if tested cell is hidden
/////////////////////////////////////////////////////////////////////////////
inline BOOL testCell (int x, int y, int relative_x, int relative_y)
{
	if ((x < WIDTH) && (y < HEIGHT) && (x >= 0) && (y >= 0))
	{
		int y_flipped = (HEIGHT-1)-y;

		int offset = MK_OFFSET (x, y_flipped); 

		if (!H[offset]) // cell not hidden
		{
			// cell of TBVs matrix not empty
			if (!T[offset].empty ())
				add2PVSTBV (x, y_flipped);

			if (!I[offset].empty()) // cell of ID matrix not empty
			{
				// add objects of this cell to the PVS
				add2PVS (x, y_flipped, FALSE);

				if (!flagInsideObj && O[offset]) // is it an occluder?
				{
					// extend the occluder and set up the occluder shaft
					setOccluderShaft (x, y_flipped, relative_x, relative_y);
					occluders++;
				}
			}
		}
		else
			return TRUE;
	}

	return FALSE;
} // testCell

/////////////////////////////////////////////////////////////////////////////
// Used by startTraversal()
/////////////////////////////////////////////////////////////////////////////
void traversalAux (int x_ini, int y_ini, int x, int y, int orx, int ory)
{
	return;

	int x1 = x_ini + orx,
		y1 = (HEIGHT - 1) - (y_ini + ory),
		x2 = x + orx,
		y2 = (HEIGHT - 1) - (y + ory);
					
	int vPosY_flipped = (HEIGHT - 1) - vPosY;

	poly.SetXY (0, x1, y1);
	poly.SetXY (1, x2, y2);
	poly.SetXY (2, vPosX+(x2-vPosX)*1000, 
				vPosY_flipped+(y2-vPosY_flipped)*1000);
	poly.SetXY (3, vPosX+(x1-vPosX)*1000, 
				vPosY_flipped+(y1-vPosY_flipped)*1000);
	poly.SetSize (4);

	drawShaft ();
} // traversalAux

/////////////////////////////////////////////////////////////////////////////
// Start view frustum traversal from a given point of the line of sight
/////////////////////////////////////////////////////////////////////////////
void startTraversal (int orx, int ory, int x, int y)
{
	BOOL lhidden; // TRUE when a line is hidden

	if(flagInsideObj) return;

	int x_ini = x, y_ini = y; // original xy
	int x_dest, y_dest; // relative to the viewpoint

	if (abs(x) > abs(y)) // up and down
	{
		if (x > 0) // at right
		{
			// go up
			if (goUp)
			{
				lhidden = TRUE;
				if (!isInsideVF (x+orx, -x+ory))
				{
					// determine boundary cell below (x,-x) and go straight up to that cell
					y_dest = -(((al*(x+orx))+cl)/bl) - ory;
					do {  
						lhidden &= testCell (x+orx, y+ory, x, y); 
						y--; 
					} while (y >= y_dest);
				}
				else
				{
					// go straight up to (x,-x), search boundary cell to the left and go straight left to that cell
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y);
						y--; 
					} while (y > -x);

					if (goLeft)
					{
						x_dest = -(((bl*(y+ory))+cl)/al) - orx;
						do { 
							lhidden &= testCell (x+orx, y+ory, x, y); 
							x--; 
						} while (x >= x_dest);
					}
				}
				if (lhidden) 
				{
					traversalAux (x_ini, y_ini, x, y, orx, ory);
					goUp = FALSE;
				}
			}

			x = x_ini; y = y_ini; // reset xy coordinates

			// go down
			if (goDown)
			{
				lhidden = TRUE;
				if (!isInsideVF (x+orx, x+ory))
				{
					// determine boundary cell above (x,x) and go straight down to that cell
					y_dest = -(((ar*(x+orx))+cr)/br) - ory;
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						y++; 
					} while (y <= y_dest);
				}
				else
				{
					// go straight down to (x,x), search boundary cell to the left and go straight left to that cell
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						y++; 
					} while (y < x);
					
					if (goLeft)
					{
						x_dest = -(((br*(y+ory))+cr)/ar) - orx;
						do { 
							lhidden &= testCell (x+orx, y+ory, x, y); 
							x--; 
						} while (x >= x_dest);
					}
				}
				if (lhidden)
				{
					traversalAux (x_ini, y_ini, x, y, orx, ory);
					goDown = FALSE;
				}
			} 
		}
		else // at left
		{
			// go up
			if (goUp)
			{
				lhidden = TRUE;
				if (!isInsideVF (x+orx, x+ory))
				{
					// determine boundary cell below (x,x) and go straight up to that cell
					y_dest = -(((ar*(x+orx))+cr)/br) - ory;
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						y--; 
					} while (y >= y_dest);
				}
				else
				{
					// go straight up to (x,x), search boundary cell to the right and go straight right to that cell
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						y--; 
					} while (y > x);
					
					if (goRight)
					{
						x_dest = -(((br*(y+ory))+cr)/ar) - orx;
						do { 
							lhidden &= testCell (x+orx, y+ory, x, y); 
							x++; 
						} while (x <= x_dest);
					}
				}
				if (lhidden) 
				{
					traversalAux (x_ini, y_ini, x, y, orx, ory);
					goUp = FALSE;
				}
			}

			x = x_ini; y = y_ini; // reset xy coordinates

			// go down
			if (goDown)
			{
				lhidden = TRUE;
				if (!isInsideVF (x+orx, -x+ory))
				{
					// determine boundary cell above (x,-x) and go straight down to that cell
					y_dest = -(((al*(x+orx))+cl)/bl) - ory;
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						y++; 
					} while (y <= y_dest);
				}
				else
				{
					// go straight down to (x,-x), search boundary cell to the right and go straight right to that cell
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						y++; 
					} while (y < -x);

					if (goRight)
					{
						x_dest = -(((bl*(y+ory))+cl)/al) - orx;
						do { 
							lhidden &= testCell (x+orx, y+ory, x, y); 
							x++; 
						} while (x <= x_dest);
					}
				}
				if (lhidden) 
				{
					traversalAux (x_ini, y_ini, x, y, orx, ory);
					goDown = FALSE;
				}
			}
		}
	}
	else if (abs(x) < abs(y)) // left and right
	{
		if (y > 0) // at down
		{
			// go right
			if (goRight)
			{
				lhidden = TRUE;
				if (!isInsideVF (y+orx, y+ory))
				{
					// determine boundary cell to the left of (y,y) and go straight right to that cell
					x_dest = -(((bl*(y+ory))+cl)/al) - orx;
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						x++; 
					} while (x <= x_dest);
				}
				else
				{
					// go straight right to (y,y), search boundary cell to above and go straight up to that cell
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						x++; 
					} while (x < y);

					if (goUp)
					{
						y_dest = -(((al*(x+orx))+cl)/bl) - ory;
						do { 
							lhidden &= testCell (x+orx, y+ory, x, y); 
							y--; 
						} while (y >= y_dest);
					}
				}
				if (lhidden) 
				{
					traversalAux (x_ini, y_ini, x, y, orx, ory);
					goRight = FALSE;
				}
			}

			x = x_ini; y = y_ini; // reset xy coordinates

			// go left
			if (goLeft)
			{
				lhidden = TRUE;
				if (!isInsideVF (-y+orx, y+ory))
				{
					// determine boundary to the right of (-y,y) and go straight left to that cell
					x_dest = -(((br*(y+ory))+cr)/ar) - orx;
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						x--; 
					} while (x >= x_dest);
				}
				else
				{
					// go straight left to (-y,y), search boundary cell to above and go straight up to that cell
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						x--; 
					} while (x > -y);

					if (goUp)
					{
						y_dest = -(((ar*(x+orx))+cr)/br) - ory;
						do { 
							lhidden &= testCell (x+orx, y+ory, x, y); 
							y--; 
						} while (y >= y_dest);
					}
				}
				if (lhidden) 
				{
					traversalAux (x_ini, y_ini, x, y, orx, ory);
					goLeft = FALSE;
				}
			}
		}
		else // at up
		{
			// go right
			if (goRight)
			{
				lhidden = TRUE;
				if (!isInsideVF (-y+orx, y+ory))
				{
					// determine boundary cell to the left of (-y,y) and go straight right to that cell
					x_dest = -(((br*(y+ory))+cr)/ar) - orx;
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						x++; 
					} while (x <= x_dest);
				}
				else
				{
					// go straight right to (-y,y), search boundary cell to below and go straight down to that cell
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						x++; 
					} while (x < -y);

					if (goDown)
					{
						y_dest = -(((ar*(x+orx))+cr)/br) - ory;
						do { 
							lhidden &= testCell (x+orx, y+ory, x, y); 
							y++; 
						} while (y <= y_dest);
					}
				}
				if (lhidden) 
				{
					traversalAux (x_ini, y_ini, x, y, orx, ory);
					goRight = FALSE;
				}
			}

			x = x_ini; y = y_ini; // reset xy coordinates

			// go left
			if (goLeft)
			{
				lhidden = TRUE;
				if (!isInsideVF (y+orx, y+ory))
				{
					// determine boundary to the right of (y,y) and go straight left to that cell 
					x_dest = -(((bl*(y+ory))+cl)/al) - orx;
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						x--; 
					} while (x >= x_dest);
				}
				else
				{
					// go straight left to (y,y), search boundary cell to below and go straight down to that cell
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						x--; 
					} while (x > y);

					if (goDown)
					{
						y_dest = -(((al*(x+orx))+cl)/bl) - ory;
						do { 
							lhidden &= testCell (x+orx, y+ory, x, y); 
							y++; 
						} while (y <= y_dest);
					}
				}
				if (lhidden) 
				{
					traversalAux (x_ini, y_ini, x, y, orx, ory);
					goLeft = FALSE;
				}
			}
		} 
	}
	else if (abs(x) && abs(y))
	{
		if (x > 0 && y < 0)			// 1st quadrant
		{
			// go down
			if (goDown)
			{
				lhidden = TRUE;
				if (!isInsideVF (x+orx, -y+ory))
				{
					// determine boundary cell above (x,-y) and go straight down to that cell
					y_dest = -(((ar*(x+orx))+cr)/br) - ory;
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						y++; 
					} while (y <= y_dest);
				}
				else
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						y++; 
					} while (y <= x);

				if (lhidden) 
				{
					traversalAux (x_ini, y_ini, x, y, orx, ory);
					goDown = FALSE;
				}
			}

			x = x_ini; y = y_ini; // reset xy coordinates

			// go left
			if (goLeft)
			{
				lhidden = TRUE;
				if (!isInsideVF (-x+orx, y+ory))
				{
					// determine boundary cell right to (-x,y) and go straight left to that cell
					x_dest = -(((bl*(y+ory))+cl)/al) - orx;
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						x--; 
					} while (x >= x_dest);
				}
				else
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						x--; 
					} while (x >= y);
				
				if (lhidden) 
				{
					traversalAux (x_ini, y_ini, x, y, orx, ory);
					goLeft = FALSE;
				}
			}
		}
		else if (x < 0 && y < 0)	// 2nd quadrant
		{
			// go right
			if (goRight)
			{
				lhidden = TRUE;
				if (!isInsideVF (-x+orx, y+ory))
				{
					// determine boundary cell left to (-x,y) and go straight right to that cell
					x_dest = -(((br*(y+ory))+cr)/ar) - orx;
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						x++; 
					} while (x <= x_dest);
				}
				else
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						x++; 
					} while (x <= -y);
				if (lhidden) 
				{
					traversalAux (x_ini, y_ini, x, y, orx, ory);
					goRight = FALSE;
				}
			}

			x = x_ini; y = y_ini; // reset xy coordinates

			// go down
			if (goDown)
			{
				lhidden = TRUE;
				if (!isInsideVF (x+orx, -y+ory))
				{
					// determine boundary cell above (x,-y) and go straight down to that cell
					y_dest = -(((al*(x+orx))+cl)/bl) - ory;
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						y++; 
					} while (y <= y_dest);
				}
				else
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						y++; 
					} while (y <= -x);
				if (lhidden) 
				{
					traversalAux (x_ini, y_ini, x, y, orx, ory);
					goDown = FALSE;
				}
			}
		}
		else if (x < 0 && y > 0)	// 3rd quadrant
		{
			// go right
			if (goRight)
			{
				lhidden = TRUE;
				if (!isInsideVF (-x+orx, y+ory))
				{
					// determine boundary cell left to (-x,y) and go straight right to that cell
					x_dest = -(((bl*(y+ory))+cl)/al) - orx;
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						x++; 
					} while (x <= x_dest);
				}
				else
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						x++; 
					} while (x <= y);
				if (lhidden) 
				{
					traversalAux (x_ini, y_ini, x, y, orx, ory);
					goRight = FALSE;
				}
			}

			x = x_ini; y = y_ini; // reset xy coordinates

			// go up
			if (goUp)
			{
				lhidden = TRUE;
				if (!isInsideVF (x+orx, -y+ory))
				{
					// determine boundary cell below (x,-y) and go straight up to that cell
					y_dest = -(((ar*(x+orx))+cr)/br) - ory;
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						y--; 
					} while (y >= y_dest);
				}
				else
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						y--; 
					} while (y >= x);
				if (lhidden) 
				{
					traversalAux (x_ini, y_ini, x, y, orx, ory);
					goUp = FALSE;
				}
			}
		}
		else if (x > 0 && y > 0)	// 4th quadrant
		{
			// go left
			if (goLeft)
			{
				lhidden = TRUE;
				if (!isInsideVF (-x+orx, y+ory))
				{
					// determine boundary cell right to (-x,y) and go straight left to that cell
					x_dest = -(((br*(y+ory))+cr)/ar) - orx;
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						x--; 
					} while (x >= x_dest);
				}
				else
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						x--; 
					} while (x >= -y);

				if (lhidden) 
				{
					traversalAux (x_ini, y_ini, x, y, orx, ory);
					goLeft = FALSE;
				}
			}

			x = x_ini; y = y_ini; // reset xy coordinates

			// go up
			if (goUp)
			{
				lhidden = TRUE;
				if (!isInsideVF (x+orx, -y+ory))
				{
					// determine boundary cell below (x,-y) and go straight up to that cell
					y_dest = -(((al*(x+orx))+cl)/bl) - ory;
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						y--; 
					} while (y >= y_dest);
				}
				else
					do { 
						lhidden &= testCell (x+orx, y+ory, x, y); 
						y--; 
					} while (y >= -x);
				
				if (lhidden)
				{
					traversalAux (x_ini, y_ini, x, y, orx, ory);
					goUp = FALSE;
				}
			}
		}
	}
} // startTraversal

/////////////////////////////////////////////////////////////////////////////
// Transform view frustum coordinates
/////////////////////////////////////////////////////////////////////////////
void transformVFC (void)
{
	float fovRad = (fov) * 1.74532925199e-2; // convert degrees to radians
	float rotRad = rot * 1.74532925199e-2;

	corner1X = vPosX+zFar;
	corner1Y = vPosY-(tan(fovRad/2)*zFar);

	corner2X = vPosX+zFar;
	corner2Y = vPosY+(tan(fovRad/2)*zFar);

	// transformed corner points
	tCorner1X = ( (corner1X-vPosX)*cos(rotRad) + 
				(corner1Y-vPosY)*sin(rotRad))+vPosX;
	tCorner1Y = (-(corner1X-vPosX)*sin(rotRad) + 
				(corner1Y-vPosY)*cos(rotRad))+vPosY;
	tCorner2X = ( (corner2X-vPosX)*cos(rotRad) + 
				(corner2Y-vPosY)*sin(rotRad))+vPosX;
	tCorner2Y = (-(corner2X-vPosX)*sin(rotRad) + 
				(corner2Y-vPosY)*cos(rotRad))+vPosY;

	// clipping line coefficients
	al =   tCorner1Y-vPosY;	
	bl = -(tCorner1X-vPosX);
	cl = -vPosX*al+vPosY*-bl;

	ar =   vPosY-tCorner2Y;	
	br = -(vPosX-tCorner2X);
	cr = -tCorner2X*ar+tCorner2Y*-br;
} // transformVFC

/////////////////////////////////////////////////////////////////////////////
// Traverse the voxels spanned by a line (x1,y1)-(x2,y2) using Bresenham
/////////////////////////////////////////////////////////////////////////////
void line2D (int x1, int y1, int x2, int y2)
{
	goUp = TRUE;
	goDown = TRUE;
	goLeft = TRUE;
	goRight = TRUE;

	int x, y, dx, dy, d, incx, incy, passoy;
	int incx2;

	int orx = x1, ory = y1; // coordinates relative to the viewer

	x2 = x2-orx; y2 = y2-ory;
	x1 = x1-orx; y1 = y1-ory;
	x = x1; y = y1;
	dx = x2-x1; dy = y2-y1;

	if (x2 < 0)	{ dx = -dx;	incx2 = -1; } else incx2 = 1;
	if (dy < 0)	{ dy = -dy;	passoy = -1; } else	passoy = 1;

	if (dx >= dy)
	{
		d = (dy << 1) - dx;
		incx = dy << 1;
		incy = incx - (dx << 1);
		while (1)
		{
			startTraversal (orx, ory, x, y);

			if (x == x2)
				break;
			x += incx2;
			if (d >= 0)
			{
				y += passoy;
				d += incy;
			}
			else
				d += incx;
		 }
	}
	else
	{
		d = (dx << 1) - dy;
		incy = dx << 1;
		incx = incy - (dy << 1);
		while (1)
		{
			startTraversal (orx, ory, x, y);

			if (y == y2)
				break;
			y += passoy;
			if (d >= 0)
			{
				x += incx2;
				d += incx;
			}
			else
				d += incy;
		}
	}
} // line2D

/////////////////////////////////////////////////////////////////////////////
// Draw bitmapped text at xy location
/////////////////////////////////////////////////////////////////////////////
void printString (float x, float y, char *text, ...)
{
	char atmp[256];

	va_list marker;
	va_start (marker, text);
	vsprintf (atmp, text, marker);
	va_end (marker);

	glColor3f (1.0, 1.0, 1.0);
	glRasterPos2f (x, y);
	for(unsigned int ct = 0; ct < strlen (atmp); ct++)
		glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, atmp[ct]);
} // printString

/////////////////////////////////////////////////////////////////////////////
// Update the scene for a visible dynamic object
/////////////////////////////////////////////////////////////////////////////
void upd_vis (int ct, int draw_occ)
{
	int map_x, map_y, c_x, c_y;

	int x = (int) disks[ct].GetLastX () - disks[ct].GetRadius () - 1,
		y = (int) disks[ct].GetLastY () - disks[ct].GetRadius () - 1;
	int w, h;
	w = h = (disks[ct].GetRadius () * 2) + 3;

	// remove object of the last position
	for (map_y = h-1, c_y = HEIGHT-y-h; map_y >= 0; map_y--, c_y++)
	{
		for (map_x = 0, c_x = x; map_x < w; map_x++, c_x++)
			// reset labels matrix
			I[MK_OFFSET (c_x, c_y)].erase (ct);
	}

	if (!draw_occ)
		return;

	// insert object in the new position
	x = (int) disks[ct].GetX () - disks[ct].GetRadius () - 1;
	y = (int) disks[ct].GetY () - disks[ct].GetRadius () - 1;

	disks[ct].SetLastX (disks[ct].GetX ());
	disks[ct].SetLastY (disks[ct].GetY ());
	
	disks[ct].DrawDisk ();

	glReadPixels (x, y, w, h, GL_RED, GL_FLOAT, readBuffer);

	glColor3f (0.0, 0.0, 0.0);
	glBegin (GL_QUADS);
		glVertex3f (x, y, 0);
		glVertex3f (x, y + h, 0);
		glVertex3f (x + w, y + h, 0);
		glVertex3f (x + w, y, 0);
	glEnd ();

	// (HEIGHT-1)-y-(h-1)
	for (map_y = h-1, c_y = HEIGHT-y-h; map_y >= 0; map_y--, c_y++)
	{
		for (map_x = 0, c_x = x; map_x < w; map_x++, c_x++)
		{
			if (readBuffer[(map_y*w)+map_x] != 0.0)
			{
				if (readBuffer[(map_y*w)+map_x] == 1.0)
					O.set(MK_OFFSET(c_x, c_y));

				I[MK_OFFSET(c_x, c_y)].insert (ct);
			}
		}
	}
} // upd_vis

/////////////////////////////////////////////////////////////////////////////
// Update the scene for a visible dynamic object
/////////////////////////////////////////////////////////////////////////////
void upd_vis_static (int ct)
{
	// insert object in the new position
	int x = (int) disks[ct].GetX () - disks[ct].GetRadius () - 1;
	int y = (int) disks[ct].GetY () - disks[ct].GetRadius () - 1;
	int w, h;
	w = h = (disks[ct].GetRadius () * 2) + 3;

	disks[ct].DrawDisk ();

	glReadPixels (x, y, w, h, GL_RED, GL_FLOAT, readBuffer);

	glColor3f (0.0, 0.0, 0.0);
	glBegin (GL_QUADS);
		glVertex3f (x, y, 0);
		glVertex3f (x, y + h, 0);
		glVertex3f (x + w, y + h, 0);
		glVertex3f (x + w, y, 0);
	glEnd ();

	int map_x, map_y, c_x, c_y;

	// (HEIGHT-1)-y-(h-1)
	for (map_y = h-1, c_y = HEIGHT-y-h; map_y >= 0; map_y--, c_y++)
	{
		for (map_x = 0, c_x = x; map_x < w; map_x++, c_x++)
		{
			if (readBuffer[(map_y*w)+map_x] != 0.0)
			{
				if (readBuffer[(map_y*w)+map_x] == 1.0)
					O_static.set(MK_OFFSET(c_x, c_y));

				I[MK_OFFSET(c_x, c_y)].insert (ct);
			}
		}
	}
} // upd_vis_static

/////////////////////////////////////////////////////////////////////////////
// Setup structures and state variables
/////////////////////////////////////////////////////////////////////////////
void init (void) 
{
	long ct;

	glClearColor (0.0, 0.0, 0.0, 0.0);
	glShadeModel (GL_FLAT);

	qobj = gluNewQuadric ();

	if (!(readBuffer = new float[WIDTH*HEIGHT]))
	{
		fprintf (stderr, "Not enough memory.\n");
		exit (-1);
	}

	if (!(readBufferR = new float[WIDTH+1]) ||
		!(readBufferG = new float[WIDTH+1]) ||
		!(readBufferB = new float[WIDTH+1]))
	{
		fprintf (stderr, "Not enough memory.\n");
		exit (-1);
	}

	// initialize disks
	if (!(disks = new Disk[N_DISKS_DYN + N_DISKS_STC]))
	{
		fprintf (stderr, "Not enough memory.\n");
		exit (-1);
	}

	float minRadius = (SQUARE(WIDTH, HEIGHT)*3)/SQUARE(256, 256);
	float maxRadius = (SQUARE(WIDTH, HEIGHT)*4)/SQUARE(256, 256);

	n_disks = N_DISKS_DYN + N_DISKS_STC;

	for (ct = 0; ct < n_disks; ct++)
	{
		disks[ct].SetSides (11); // 5
		disks[ct].SetRadius (minRadius + (rand() % (int)maxRadius));

		if (ct >= N_DISKS_DYN)
			disks[ct].SetVelocity (0.0);
		else
			disks[ct].SetVelocity (-.05 + (2.0 / (1 + (rand() % 50))));

		disks[ct].SetBounds (WIDTH, HEIGHT);
		disks[ct].SetPos (disks[ct].GetRadius() + 
			(4 + rand() % (int)(WIDTH - (disks[ct].GetRadius()*2) - 8)),
			disks[ct].GetRadius() + 
			(4 + rand() % (int)(HEIGHT - (disks[ct].GetRadius()*2) - 8)));

		// TBV settings
		disks[ct].SetTBVX (disks[ct].GetX ());
		disks[ct].SetTBVY (disks[ct].GetY ());
		disks[ct].SetStartDate (current_date);
		disks[ct].SetExpDate (current_date + 5 + (rand() % 40));
		disks[ct].SetLastX (disks[ct].GetX ());
		disks[ct].SetLastY (disks[ct].GetY ());
		disks[ct].SetTBVState (0); // TBV not assigned

		R[ct] = 1.0 / (rand() % 10);
		G[ct] = 1.0 / (rand() % 10);
		B[ct] = 1.0 / (rand() % 10);

		PVS[ct] = 1; // everything's visible
	}

	for (ct = 0; ct < WIDTH*HEIGHT; ct++)
	{
		T[ct].clear ();
		I[ct].clear ();
	}

	O_static.reset ();
	O.reset ();
	H.reset ();

	resetFPS ();
}

void reshape (int w, int h)
{
	glViewport (0, 0, (GLsizei) w, (GLsizei) h);
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

	gluOrtho2D (0, w, 0, h);

	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	glTranslatef (0.375, 0.375, 0.0);
}

void perspective (int window)
{
	switch (window)
	{
	case 0:
		glViewport (0, HEIGHT+25, WIDTH, 25);
		glMatrixMode (GL_PROJECTION);
		glLoadIdentity ();
		gluPerspective (fov*(25.0/WIDTH), WIDTH/25.0, 0.5, 10000);

		glMatrixMode (GL_MODELVIEW);
		glLoadIdentity ();
		gluLookAt (vPosX, 0, vPosY, (tCorner1X+tCorner2X)/2, 0, 
				  (tCorner1Y+tCorner2Y)/2, 0, 1, 0);
	break;
	case 1:
		glViewport (0, HEIGHT, WIDTH, 25);
		glMatrixMode (GL_PROJECTION);
		glLoadIdentity ();
		gluPerspective (fov*(25.0/WIDTH), WIDTH/25.0, 0.5, 10000);

		glMatrixMode (GL_MODELVIEW);
		glLoadIdentity ();
		gluLookAt (vPosX, 0, vPosY, (tCorner1X+tCorner2X)/2, 0, 
				  (tCorner1Y+tCorner2Y)/2, 0, 1, 0);
	break;
	}
}

void display (void)
{
	long ct;
	static long logFPSCounter = 99; // 99

	static BOOL firstFrame = TRUE;

	if(firstFrame)
	{
		// discretize static visible objects once 
		// (static disks are allocated above dynamic objetcs)
		for (ct = N_DISKS_DYN; ct < N_DISKS_DYN+N_DISKS_STC; ct++)
			upd_vis_static (ct);

		O.getFromBitset(O_static);

		firstFrame = FALSE;
	}

#ifdef NHOC

	glClearColor (0.0, 0.0, 0.0, 0.0);
	glClear (GL_COLOR_BUFFER_BIT);

	// discretize last visible objects
	for (ct = 0; ct < N_DISKS_DYN; ct++)
	{
		if (PVS[ct])
		{
			// object doesn't have a TBV
			upd_vis (ct, 1);
		}
		else
		{
			// does the object have a TBV?
			if (disks[ct].HasTBV ())
			{
				// TBV expired
				if (disks[ct].GetExpDate () <= current_date)
				{
					long last_interval = disks[ct].GetExpDate () - disks[ct].GetStartDate ();

					// assign a new TBV
					eraseTBV (ct);

					disks[ct].SetTBVX (disks[ct].GetX ());
					disks[ct].SetTBVY (disks[ct].GetY ());
					disks[ct].SetStartDate (current_date);
					disks[ct].SetExpDate (current_date + ((last_interval+1) * 1.3));
					insertTBV (ct);

					disks[ct].SetTBVState (1);
				}
			}
			else
			{
				long last_interval = disks[ct].GetExpDate () - disks[ct].GetStartDate ();

				// assign a TBV
				// object is just getting invisible
				disks[ct].SetTBVX (disks[ct].GetX ());
				disks[ct].SetTBVY (disks[ct].GetY ());
				disks[ct].SetStartDate (current_date);
				disks[ct].SetExpDate (current_date + (last_interval / 2.0));
				insertTBV (ct);
				
				upd_vis (ct, 0);

				disks[ct].SetTBVState (1);
			}
		}
	}

	transformVFC ();

	flagInsideObj = 0;

	if (O[MK_OFFSET(vPosX, (HEIGHT-1)-vPosY)] ||
		O_static[MK_OFFSET(vPosX, (HEIGHT-1)-vPosY)])
		flagInsideObj = 1;

	// traverse discretized line of sight
	if (!O_static[MK_OFFSET(vPosX, (HEIGHT-1)-vPosY)])
		line2D (vPosX, vPosY, (tCorner1X+tCorner2X)/2, (tCorner1Y+tCorner2Y)/2);
	else
		add2PVS (vPosX, (HEIGHT-1)-vPosY, FALSE);

#else

	transformVFC ();

#endif

	glClearColor (1.0, 1.0, 1.0, 0.0);
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// show hidden cells
	if (show_data)
	{
		glBegin (GL_POINTS);
		for (int y = 0; y < HEIGHT; y++)
		{
			for (int x = 0; x < WIDTH; x++)
			{
				if (H[MK_OFFSET (x, (HEIGHT-1)-y)])
				{
					glColor3f ((float)0.7, (float)0.7, 1.0);
					glVertex3f (x, y, 0.0);
				}
				if (!T[MK_OFFSET (x, (HEIGHT-1)-y)].empty ())
				{
					glColor3f (0.0, 0.0, 0.0);
					glVertex3f (x, y, 0.0);
				}
				if (I[MK_OFFSET (x, (HEIGHT-1)-y)].size ())
				{
					glColor3f (0.5, 0.0, 0.0);
					glVertex3f (x, y, 0.0);
				}
				if (O[MK_OFFSET (x, (HEIGHT-1)-y)])
				{
					glColor3f (1.0, 0.0, 0.0);
					glVertex3f (x, y, 0.0);
				}
			}
		}
		glEnd ();
	}

	H.reset ();
	O.reset ();
	O.getFromBitset(O_static);

	// determine number of visible objects
	int visObjs = 0;
	for (ct = 0; ct < n_disks; ct++)
	{
		if(PVS[ct]) 
			visObjs++;
	}

	double FPS = getFPS ();

	glLineWidth (1.0);
	glColor3f (0.0, 0.0, 1.0);
	glBegin (GL_LINES);
		glVertex3f (vPosX, vPosY, 0.0);
		glVertex3f (tCorner1X, tCorner1Y, 0.0);
		glVertex3f (vPosX, vPosY, 0.0);
		glVertex3f (tCorner2X, tCorner2Y, 0.0);
	glEnd ();

	glLineStipple (2, 0xAAAA);
	glEnable (GL_LINE_STIPPLE);

	glBegin (GL_LINES);
		glVertex3f (vPosX, vPosY, 0.0);
		glVertex3f ((tCorner1X+tCorner2X)/2, (tCorner1Y+tCorner2Y)/2, 0.0);
	glEnd ();

	glDisable (GL_LINE_STIPPLE);

	// draw viewpoint
	glColor3f (0.0, 0.0, 0.0);
	glBegin (GL_POINTS);
		glVertex3f (vPosX, vPosY, 0.0);
	glEnd ();

	// clear text window
	glColor3f (0.0, 0.0, 0.0);
	glBegin (GL_QUADS);
		glVertex3f (    0,  HEIGHT+50, 0);
		glVertex3f (    0, HEIGHT+100, 0);
		glVertex3f (WIDTH, HEIGHT+100, 0);
		glVertex3f (WIDTH,  HEIGHT+50, 0);
	glEnd ();

	// draw scene from the viewpoint
	// clear scene window
	glColor3f (0.0, 0.0, 0.0);
	glBegin (GL_QUADS);
		glVertex3f (    0,    HEIGHT, 0);
		glVertex3f (    0, HEIGHT+50, 0);
		glVertex3f (WIDTH, HEIGHT+50, 0);
		glVertex3f (WIDTH,    HEIGHT, 0);
	glEnd ();

#ifdef NHOC

	glEnable (GL_CULL_FACE);
	glEnable (GL_DEPTH_TEST);

	perspective (0);

	// draw potentially visible cylinders
	for (ct = 0; ct < n_disks; ct++)
	{
		if(PVS[ct]) 
		{
			glPushMatrix ();
				glColor3f (R[ct], G[ct], B[ct]);
				glTranslatef (disks[ct].GetX (), 100.0, disks[ct].GetY ());
				glRotatef (90, 1.0, 0.0, 0.0);
				gluCylinder (qobj, disks[ct].GetRadius (), disks[ct].GetRadius (), 500, 40, 12);
			glPopMatrix ();
		}
	}

	// clear scene window
	reshape (WIDTH, HEIGHT+100);

	glDisable (GL_CULL_FACE);
	glDisable (GL_DEPTH_TEST); 

	glReadPixels (0, HEIGHT+26, WIDTH, 1, GL_RED, GL_FLOAT, readBufferR);
	glReadPixels (0, HEIGHT+26, WIDTH, 1, GL_GREEN, GL_FLOAT, readBufferG);
	glReadPixels (0, HEIGHT+26, WIDTH, 1, GL_BLUE, GL_FLOAT, readBufferB);

	float objCurR = 0.0, objCurG = 0.0, objCurB = 0.0;
	int imgVisObjs = 0;
	for (int cx = 0; cx < WIDTH; cx++)
	{
		if (readBufferR[cx] || readBufferG[cx] || readBufferB[cx])
		{
			if ((objCurR != readBufferR[cx]) || (objCurG != readBufferG[cx]) || 
				(objCurB != readBufferB[cx]))
			{
				objCurR = readBufferR[cx];
				objCurG = readBufferG[cx];
				objCurB = readBufferB[cx];
				imgVisObjs++;
			}
		}
	}

	if(imgVisObjs)
		printString (5, HEIGHT+0+8, "Tightness: %.2f", (float)visObjs/(float)imgVisObjs);
	else
		printString (5, HEIGHT+0+8, "Tightness: %.2f", 1.0);

	float tightness = 0;

#else

	glColor3f (0.0, 0.0, 0.0);
	glBegin (GL_QUADS);
		glVertex3f (    0,    HEIGHT, 0);
		glVertex3f (    0, HEIGHT+25, 0);
		glVertex3f (WIDTH, HEIGHT+25, 0);
		glVertex3f (WIDTH,    HEIGHT, 0);
	glEnd ();
	
	glEnable (GL_CULL_FACE);
	glEnable (GL_DEPTH_TEST);

	perspective (1);

	// draw visible and invisible cylinders 
	for (ct = 0; ct < n_disks; ct++)
	{
		glPushMatrix ();
			glColor3f (R[ct], G[ct], B[ct]);
			glTranslatef (disks[ct].GetX (), 100.0, disks[ct].GetY ());
			glRotatef (90, 1.0, 0.0, 0.0);
			gluCylinder (qobj, disks[ct].GetRadius (), disks[ct].GetRadius (), 500, 40, 12);
		glPopMatrix ();
	}

	glDisable (GL_CULL_FACE);
	glDisable (GL_DEPTH_TEST);

	// back to original projection
	reshape (WIDTH, HEIGHT+100);

#endif

	glColor3f (1.0, 1.0, 1.0);
	glBegin (GL_LINES);
		glVertex3f (    0, HEIGHT+24, 0);
		glVertex3f (WIDTH, HEIGHT+24, 0);
		glVertex3f (    0, HEIGHT+49, 0);
		glVertex3f (WIDTH, HEIGHT+49, 0);
	glEnd ();

	// draw text
	printString (5, HEIGHT+50+5, "FPS: %.1f  Frame #%d", FPS, current_date);
	printString (5, HEIGHT+50+20, "X: %.1f  Y: %.1f  Rot: %.1f  FOV: %.1f", vPosX, vPosY, rot, fov);
	printString (5, HEIGHT+50+35, "Conservative set: %d  Exact set: %d  Occluders: %d", visObjs, imgVisObjs, occluders);

	if (show_data)
	{
		float overwrite = ((overwrite_cur - overwrite_min) / (overwrite_max - overwrite_min)) * 100.0;

		if (!(overwrite_max - overwrite_min))
			overwrite = 0;
	}

	// write FPS to log file
	if(current_date >= 100 && current_date <= 1100) // 100, 1100
	{
		logFPSCounter++;
		if (logFPSCounter == 100) // 100
		{
			AddToLog (FPS, visObjs, occluders, tightness);
			logFPSCounter = 0;
		}
	}

	occluders = 0;

	// next frame
	if (is_dynamic)
	{
		for (ct = 0; ct < n_disks; ct++)
			disks[ct].Move ();
	}

	memset (PVS, 0, n_disks); // clear PVS

	current_date++;

	glutSwapBuffers ();
	glFinish ();
}

void keyboard (unsigned char key, int x, int y)
{
	switch (key) {
	case GLUT_KEY_LEFT:
		if (glutGetModifiers () == GLUT_ACTIVE_SHIFT)
		{
			if (vPosX > 10) vPosX-=10;
		}
		else
			if (vPosX > 0) vPosX--;
		break;

	case GLUT_KEY_RIGHT:
		if (glutGetModifiers () == GLUT_ACTIVE_SHIFT)
		{
			if (vPosX < WIDTH-11) vPosX+=10;
		}
		else
			if (vPosX < WIDTH-1) vPosX++;
		break;

	case GLUT_KEY_UP:
		if (glutGetModifiers () == GLUT_ACTIVE_SHIFT)
		{
			if (vPosY > 10) vPosY-=10;
		}
		else
			if (vPosY > 0) vPosY--;
		break;

	case GLUT_KEY_DOWN:
		if (glutGetModifiers () == GLUT_ACTIVE_SHIFT)
		{
			if (vPosY < HEIGHT-11) vPosY+=10;
		}
		else
			if (vPosY < HEIGHT-1) vPosY++;
		break;

	case GLUT_KEY_HOME:
		(rot < 359) ? rot++ : rot = 0;
		break;

	case GLUT_KEY_END:
		(rot > 0) ? rot-- : rot = 359;
		break;

	case '1':
		if (fov < 179) fov++;
		break;
	
	case '2':
		if (fov > 0) fov--;
		break;

	case '3':
		show_data = !show_data;
		break;

	case '4':
		is_dynamic = !is_dynamic;
		break;

	case 27: // esc
		WriteLogFile ();

		gluDeleteQuadric (qobj);
		if (disks) delete []disks;
		if (readBuffer) delete []readBuffer;
		if (readBufferR) delete []readBufferR;
		if (readBufferG) delete []readBufferG;
		if (readBufferB) delete []readBufferB;
		exit(0);
		break;
	}

	glutPostRedisplay ();
}

void idle (void)
{
	glutPostRedisplay ();
}

int main (int argc, char** argv)
{
	glutInit (&argc, argv);
	glutInitDisplayMode (GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGBA);
	glutInitWindowSize (WIDTH, HEIGHT + 50 + 50); 
	glutCreateWindow (argv[0]);
	glutSetWindowTitle ("2D dynamic visibility");

	printf ("2D dynamic visibility\n");
	printf ("Size: %dx%d  Objects: %d\n\n", WIDTH, 
		    HEIGHT, N_DISKS_DYN+N_DISKS_STC);
	printf ("Keys:\nARROWS   move the viewpoint\n");
	printf ("HOME/END rotate\n");
	printf ("1        increase field of view\n");
	printf ("2        decrease field of view\n");
	printf ("3        toggle structure display mode\n");
	printf ("4        toggle beetwen dynamic or static scene\n\n");

	init ();
	glutDisplayFunc (display); 
	glutReshapeFunc (reshape);
	glutKeyboardFunc (keyboard);
	glutIdleFunc (idle);
	glutMainLoop ();

	WriteLogFile ();

	gluDeleteQuadric (qobj);
	delete []disks;
	if (readBufferR) delete []readBufferR;
	if (readBufferG) delete []readBufferG;
	if (readBufferB) delete []readBufferB;

	return 0;
}