//***************************************************************************
//
// POLY2D.CPP
// by Harlen C. Batagelo
//
//***************************************************************************

#include "poly2d.h"

/////////////////////////////////////////////////////////////////////////////
// Clip polygon against a rectangular boundary
/////////////////////////////////////////////////////////////////////////////
void Poly2D::ClipRect (float left, float top, float right, float bottom)
{
	Vertex2D a[MAX_VERTEX], b[MAX_VERTEX];
	
	float x0, y0, x1, y1, x_i, y_i;
	int curr, last;
	int n = 0;

	// clip against left
	last = size-1;
	for (curr = 0; curr < size; curr++)
	{
		x1 = GetX (curr%size), y1 = GetY (curr%size);
		x0 = GetX (last), y0 = GetY (last);

		if ((x1 >= left) && (x0 >= left))
			a[n++].SetXY (x1, y1); // inside
		else if (!((x1 < left) && (x0 < left))) // not completely outside
		{
			y_i = y0+(y1-y0)*((left-x0)/(x1-x0));
			a[n++].SetXY (left, y_i); // exiting

			if (x1 >= left) // entering
				a[n++].SetXY (x1, y1);
		}

		last = curr;
	}
	size = n;

	// clip against right
	n = 0;
	last = size-1;
	for (curr = 0; curr < size; curr++)
	{
		x1 = a[curr%size].GetX (), y1 = a[curr%size].GetY ();
		x0 = a[last].GetX (), y0 = a[last].GetY ();

		if ((x1 <= right) && (x0 <= right))
			b[n++].SetXY (x1, y1); // inside
		else if (!((x1 > right) && (x0 > right))) // not completely outside
		{
			y_i = y0+(y1-y0)*((right-x0)/(x1-x0));
			b[n++].SetXY (right, y_i); // exiting

			if (x1 <= right) // entering
				b[n++].SetXY (x1, y1);
		}

		last = curr;
	}
	size = n;

	// clip against bottom
	n = 0;
	last = size-1;
	for (curr = 0; curr < size; curr++)
	{
		x1 = b[curr%size].GetX (), y1 = b[curr%size].GetY ();
		x0 = b[last].GetX (), y0 = b[last].GetY (); 

		if ((y1 <= bottom) && (y0 <= bottom)) // both inside
			a[n++].SetXY (x1, y1); 
		else if (!((y1 > bottom) && (y0 > bottom))) // not completely outside
		{
			x_i = x0+(x1-x0)*((bottom-y0)/(y1-y0));
			a[n++].SetXY (x_i, bottom); // exiting

			if (y1 <= bottom) // entering
				a[n++].SetXY (x1, y1); 
		}
		last = curr;
	}
	size = n;

	// clip against top
	n = 0;
	last = size-1;
	for (curr = 0; curr < size; curr++)
	{
		x1 = a[curr%size].GetX (), y1 = a[curr%size].GetY ();
		x0 = a[last].GetX (), y0 = a[last].GetY (); 

		if ((y1 >= top) && (y0 >= top)) // both inside
			SetXY (n++, x1, y1); 
		else if (!((y1 < top) && (y0 < top))) // not completely outside
		{
			x_i = x0+(x1-x0)*((top-y0)/(y1-y0));
			SetXY (n++, x_i, top); // exiting

			if (y1 >= top) // entering
				SetXY (n++, x1, y1); 
		}
		last = curr;
	}
	size = n;
} // Poly2D::ClipRect